from django.db.models.signals import post_save
from django.dispatch import receiver

from django_infra.feature_flags.flags import retrieve_feature_flag_from_db
from django_infra.feature_flags.models import FeatureFlag


@receiver(post_save, sender=FeatureFlag)
def pop_cache(*args, **kwargs):
    retrieve_feature_flag_from_db.cache_clear()



from typing import Any, Callable, Dict, List

from django_filters import rest_framework as filters
from django_filters.rest_framework import DjangoFilterBackend
from drf_partial_response.serializers import FieldsListSerializerMixin
from drf_partial_response.views import OptimizedQuerySetBase
from drf_spectacular.utils import extend_schema, extend_schema_view
from rest_framework import serializers


class FiltersetSerializerModelMetaClass(type):
    """Auto creates filterset class if one is not specified, and augments view with `filters` attribute.
        This metaclass should work in the background to enhance the power of views without modifying existing functionality
        unless the implementation specifically attempts to make use of it.

        The primary use of this metaclass is for the enabling of the `filters` attribute in a View.

        The secondary uses are for the reduction of boilerplate code associated with serializers and Filtersets.

        Examples
        --------
    import django_extras.filters    >>> # FilteredPartialResponseModelViewSet has a metaclass which inherits `FiltersetSerializerModelMetaClass`
        >>> import django_infra
        >>> from my_app.models import TestModelRelations
        >>> from my_app.serializers import M2MModelSerializer, FKModelSerializer
        >>> class OptimizedTestView(django_infra.api.views.PaginatedViewMixin, django_infra.api.views.FilteredPartialResponseModelViewSet):
        >>>     permission_classes = []
        >>>     authentication_classes = []
        >>>     model = TestModelRelations
        >>>     related_serializers_list = [M2MModelSerializer, FKModelSerializer]
        >>>     filters = [
        >>>         django_infra.api.filters.OrderingFilter(
        >>>             name='ordering',
        >>>             fields=[
        >>>                 django_infra.api.filters.FilterField(internal='date_field',external='dateField'),
        >>>                 django_infra.api.filters.FilterField(internal='id'),
        >>>             ],
        >>>         ),
        >>>         django_infra.api.filters.IContainsFilter(
        >>>             fields=[
        >>>                 django_infra.api.filters.MultiInternalFilterField(
        >>>                     internal=['char_field','date_field'], external='search'
        >>>                 )
        >>>             ]
        >>>         ),
        >>>     ]
        In this example:
        1. Filters are anabled (see documentation on each individual filter.
        2. A serializer is auto created for this class (this can be overwritten simply by specifying one)
        3. The created/specified serializer will automatically use M2MModelSerializer and FKModelSerializer on any fields matching these models.
        Note -M2MModelSerializer and FKModelSerializer are for M2MModel and FKModel test models,
        so if this view was for something like `PatientJourneyViewSet` the equivalent FKModelSerializer would be `PatientSerializer`
        and instead of receiving a `patient_id` in the response you would get a serialized patient based on the specified related serializer.


        This meta class performs the following:
        - if `filterset_class` is not specified, this Meta class will auto create one.
        - the `filterset_class` is injected with any `filterset_mixins` specified, by default `FilterSet` is injected.
        - specified `filters` (as shown in the example) are automatically injected into the `filterser_class` thereby avoiding boilerplate.

        - if `filter_backends` is not specified, this Meta class will auto create one using DjangoFilterBackend.

        - if a `serializer_class` is not specified, this Meta class will auto create one with `__all__` fields.
        - if `related_serializers` are specified then these serializers will be automatically used in related fields matching the serializers model
        -


    """

    def __new__(cls, name, bases, attrs):
        if attrs.get("model", NotImplemented) is NotImplemented:
            return super().__new__(cls, name, bases, attrs)
        if "queryset" not in attrs:
            attrs["queryset"] = attrs["model"].objects.all()
        # Precompute related_serializers mapping
        attrs["related_serializers"] = cls.precompute_related_serializers(
            attrs.get("related_serializers_list", [])
        )
        cls.create_serializer_class(attrs)
        cls.create_filterset_class(attrs)

        return super().__new__(cls, name, bases, attrs)

    @staticmethod
    def create_filterset_class(attrs: Dict[str, Any]):
        """Auto create a filterset class based on `filters` specified in view.
        Filters must be of list of type: django_extras.views.Filter
        """
        FiltersetSerializerModelMetaClass.inject_default_classes(
            attrs=attrs,
            default_classes=[
                DjangoFilterBackend,
            ],
            key="filter_backends",
        )
        FiltersetSerializerModelMetaClass.inject_default_classes(
            attrs=attrs,
            default_classes=[
                filters.FilterSet,
            ],
            key="filterset_mixins",
        )
        filters_ = attrs.get("filters", [])
        all_fields = {k: v for f in filters_ for k, v in f.to_meta().items()}
        model = attrs["model"]
        meta = type(
            "Meta",
            tuple(),
            dict(
                fields=[
                    k for k, v in all_fields.items() if not isinstance(v, Callable)
                ],
                model=attrs["model"],
            ),
        )
        kwargs = {**all_fields, "Meta": meta}
        attrs["filterset_class"] = type(
            f"{model.__name__}FilterSet", attrs.get("filterset_mixins"), kwargs
        )
        return attrs["filterset_class"]

    @staticmethod
    def precompute_related_serializers(related_serializers_list):
        return {
            serializer.Meta.model.__name__: serializer
            for serializer in related_serializers_list
        }

    @staticmethod
    def create_serializer_class(attrs: dict) -> serializers.ModelSerializer:
        model = attrs["model"]
        FiltersetSerializerModelMetaClass.inject_default_classes(
            attrs=attrs,
            default_classes=[
                FieldsListSerializerMixin,
                serializers.ModelSerializer,
            ],
            key="serializer_mixins",
        )
        field_serializers = FiltersetSerializerModelMetaClass.get_field_serializers(
            model, attrs["related_serializers"]
        )

        meta = type("Meta", tuple(), dict(fields="__all__", model=attrs["model"]))
        serializer_attrs = {"Meta": meta, **field_serializers}
        if attrs.get("serializer_class") is not None:
            attrs["serializer_mixins"] = (
                attrs.get("serializer_class"),
                *attrs["serializer_mixins"],
            )
            serializer_attrs["Meta"] = attrs.get("serializer_class").Meta
        # Inject the generated serializer class into attrs without return
        attrs["serializer_class"] = type(
            f"{model.__name__}Serializer",
            attrs.get("serializer_mixins"),
            serializer_attrs,
        )
        return attrs["serializer_class"]

    @staticmethod
    def inject_default_classes(attrs: dict, default_classes: List[object], key: str):
        """Inject default classes into an attrs key if they do not exist."""
        existing_classes = attrs.get(key, tuple())
        missing_classes = []
        for c in default_classes:
            if c not in existing_classes:
                missing_classes.append(c)
        attrs[key] = (*missing_classes, *existing_classes)

    @staticmethod
    def get_field_serializers(model, related_serializers):
        field_serializers = {}
        for field in model._meta.get_fields():
            if hasattr(field, "related_model") and field.related_model:
                related_model_name = field.related_model.__name__
                serializer = related_serializers.get(related_model_name)
                if serializer:
                    field_serializers[field.name] = serializer(
                        many=field.many_to_many or field.one_to_many
                    )
        return field_serializers


class BestGuessCrudSchemaMeta(type):
    def __new__(mcs, name, bases, namespace):
        cls = super().__new__(mcs, name, bases, namespace)
        if hasattr(cls, "model") and hasattr(cls, "serializer_class"):
            model_name = cls.model.__name__
            serializer_class = getattr(cls, "serializer_class")
            http_method_names = getattr(cls, "http_method_names", list())
            schemas = dict()
            if "get" in http_method_names:
                schemas.update(
                    list=extend_schema(
                        summary=f"List {model_name}",
                        description=f"Retrieve a list of {model_name}.",
                        responses={200: serializer_class(many=True)},
                    ),
                    retrieve=extend_schema(
                        summary=f"Retrieve {model_name}",
                        description=f"Retrieves a {model_name} by pk.",
                        responses={200: serializer_class},
                    ),
                )
            if "post" in http_method_names:
                schemas.update(
                    update=extend_schema(
                        summary=f"Update {model_name}",
                        description=f"Updates a {model_name} by pk.",
                        responses={200: serializer_class},
                    ),
                    partial_update=extend_schema(
                        summary=f"Patch {model_name}",
                        description=f"Partially updates a {model_name} by pk.",
                        responses={200: serializer_class},
                    ),
                    create=extend_schema(
                        summary=f"Create {model_name}",
                        description=f"Creates a new {model_name}.",
                        responses={201: serializer_class},
                        request=serializer_class,
                    ),
                )
            if "delete" in http_method_names:
                schemas.update(
                    destroy=extend_schema(
                        summary=f"Delete {model_name}",
                        description=f"Deletes a {model_name} by pk.",
                        responses={404: "Not Found"},
                    ),
                )
            cls = extend_schema_view(**schemas)(cls)
        return cls


class FilteredPartialResponseModelViewSetMetaClass(
    BestGuessCrudSchemaMeta, FiltersetSerializerModelMetaClass, OptimizedQuerySetBase
): ...



from rest_framework import status
from rest_framework.decorators import action
from rest_framework.permissions import SAFE_METHODS, IsAdminUser
from rest_framework.response import Response
from rest_framework.viewsets import ModelViewSet

from django_infra.feature_flags.models import FeatureFlag
from django_infra.feature_flags.serializers import FeatureFlagSerializer


class IsAdminOrReadOnlyPermission(IsAdminUser):
    def has_permission(self, request, view):
        return request.method in SAFE_METHODS or super().has_permission(request, view)


class FeatureFlagViewSet(ModelViewSet):
    queryset = FeatureFlag.objects.all()
    serializer_class = FeatureFlagSerializer
    permission_classes = [IsAdminOrReadOnlyPermission]

    def get_queryset(self):
        # hack to ensure sync occurs at least once (on access)
        # won't sync on every call.
        return FeatureFlag.objects.all()

    @action(detail=False, methods=["post"], url_path="reset")
    def reset(self, request):
        FeatureFlag.objects.reset_env_flags()
        return Response(dict(), status=status.HTTP_200_OK)

    @action(detail=True, methods=["post"], url_path="activate")
    def activate(self, request, pk=None):
        flag = self.get_object()
        flag.update(active=True)
        return Response(
            self.serializer_class(instance=flag).data, status=status.HTTP_200_OK
        )

    @action(detail=True, methods=["post"], url_path="deactivate")
    def deactivate(self, request, pk=None):
        flag = self.get_object()
        flag.update(active=False)
        return Response(
            self.serializer_class(instance=flag).data, status=status.HTTP_200_OK
        )



# Generated by Django 5.1.6 on 2025-02-23 14:28

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='QueryExport',
            fields=[
                ('id', models.CharField(max_length=64, primary_key=True, serialize=False, unique=True)),
                ('state', models.CharField(max_length=16)),
                ('metadata', models.JSONField()),
                ('file', models.FileField(upload_to='')),
                ('format', models.CharField(max_length=16)),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



from __future__ import annotations

import logging

from django.conf import settings
from django.db import models as dm

logger = logging.getLogger(__name__)


class TimeTrackingModel(dm.Model):
    created_time = dm.DateTimeField(auto_now_add=True, null=True)
    modified_time = dm.DateTimeField(auto_now=True, null=True)

    class Meta:
        abstract = True


class UserTrackingMixin(dm.Model):
    created_by = dm.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=dm.PROTECT,
        related_name="created_by_%(app_label)s_%(class)s_related",
        editable=False,
        null=True,
    )
    modified_by = dm.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=dm.PROTECT,
        related_name="modified_by_%(app_label)s_%(class)s_related",
        editable=False,
        null=True,
    )

    def save(self, *args, **kwargs):
        """Best effort attempt to set modified_by/created_by using crum."""
        try:
            from crum import get_current_user
        except ImportError:
            return super().save(*args, **kwargs)

        creating = self.id is None
        a_user_is_null = None in {self.created_by, self.modified_by}
        current_user = get_current_user()
        # attempt resolve any None fields
        if a_user_is_null and current_user:
            self.created_by = self.created_by or current_user
            self.modified_by = self.modified_by or current_user
        # best effort to update modifying user.
        if not creating and current_user != self.modified_by:
            self.modified_by = self.modified_by or current_user
        if a_user_is_null:
            logger.info("Creator or modifier is still null: %s", self)

        return super().save(*args, **kwargs)

    class Meta:
        abstract = True


class TrackingMixin(TimeTrackingModel, UserTrackingMixin):
    class Meta:
        abstract = True



# Generated by Django 5.1.6 on 2025-02-23 16:20

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('exporter', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='queryexport',
            name='state',
            field=models.CharField(choices=[('scheduled', 'Scheduled'), ('processing', 'Processing'), ('success', 'Success'), ('fail', 'Fail')], max_length=16),
        ),
    ]



from django_infra.db.models.schedule import (
    LookupOverrideMixin,
    PeriodScheduleMixin,
    ScheduleCode,
    ScheduleCodeCase,
    ScheduleMixin,
)
from django_infra.db.models.tracking import (
    TimeTrackingModel,
    TrackingMixin,
    UserTrackingMixin,
)
from django_infra.db.models.updatable import UpdatableModel

__all__ = [
    # Updatable
    "UpdatableModel",
    # Tracking
    "TimeTrackingModel",
    "UserTrackingMixin",
    "TrackingMixin",
    # Schedule
    "ScheduleCode",
    "ScheduleCodeCase",
    "ScheduleMixin",
    "PeriodScheduleMixin",
    "LookupOverrideMixin",
]



from __future__ import annotations

import abc
import datetime
import logging
from functools import partial
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from django_infra.db.querysets.schedule import ScheduleQuerySet

import pytz
from django.contrib.postgres.constraints import ExclusionConstraint
from django.contrib.postgres.fields import DateTimeRangeField, RangeOperators
from django.contrib.postgres.indexes import GistIndex
from django.db import models as dm
from django.db.models import TextChoices
from django.db.models.expressions import BaseExpression, Case, Ref, Value, When
from django.db.models.lookups import Exact, In

logger = logging.getLogger(__name__)


class ScheduleCode(TextChoices):
    # date is between start date (inclusive) and end date (exclusive)
    ACTIVE = "ACTIVE", "Active"
    # date is before start date
    FUTURE = "FUTURE", "Inactive Pre"
    # date is after end date
    EXPIRED = "EXPIRED", "Inactive Post"


class LookupOverrideMixin:
    lookup_overrides: dict[str, type] = {}

    def get_lookup(self, lookup_name):
        lookup = self.lookup_overrides.get(lookup_name)
        if lookup is not None:
            return lookup
        return super().get_lookup(lookup_name)

    @classmethod
    def unwrap_lhs(cls, lhs):
        if isinstance(lhs, cls):
            return lhs
        if isinstance(lhs, Ref) and isinstance(lhs.source, cls):
            return lhs.source
        return None

    @staticmethod
    def extract_rhs_values(rhs):
        if hasattr(rhs, "value"):
            values = rhs.value
            if isinstance(values, (list, tuple, set)):
                return list(values)
            return [values]
        if isinstance(rhs, (list, tuple, set)):
            return list(rhs)
        if not isinstance(rhs, BaseExpression):
            return [rhs]
        logger.info(
            "ScheduleCodeCase: rhs values unavailable (type=%s)", type(rhs).__name__
        )
        return None


class ScheduleCodeCase(LookupOverrideMixin, Case):
    """
    Custom Case expression for ScheduleCode that optimizes WHERE clause filtering.

    When used in SELECT (annotation), it returns the normal CASE expression.
    When used in WHERE (filtering), it converts to efficient date comparisons.
    """

    class Exact(Exact):
        def as_sql(self, compiler, connection):
            expr = ScheduleCodeCase.unwrap_lhs(self.lhs)
            if expr is None:
                logger.info(
                    "ScheduleCodeCase.Exact: not selected (lhs not ScheduleCodeCase)"
                )
                return super().as_sql(compiler, connection)
            values = ScheduleCodeCase.extract_rhs_values(self.rhs)
            if not values:
                logger.info(
                    "ScheduleCodeCase.Exact: not selected (rhs values unavailable)"
                )
                return super().as_sql(compiler, connection)
            q = expr.build_q(values[0])
            if not q:
                logger.info(
                    "ScheduleCodeCase.Exact: not selected (empty Q for %r)", values[0]
                )
                return super().as_sql(compiler, connection)
            where = q.resolve_expression(
                compiler.query, allow_joins=True, reuse=None, summarize=False
            )
            logger.info("ScheduleCodeCase.Exact: selected (optimized)")
            return compiler.compile(where)

    class In(In):
        def as_sql(self, compiler, connection):
            expr = ScheduleCodeCase.unwrap_lhs(self.lhs)
            if expr is None:
                logger.info(
                    "ScheduleCodeCase.In: not selected (lhs not ScheduleCodeCase)"
                )
                return super().as_sql(compiler, connection)
            values = ScheduleCodeCase.extract_rhs_values(self.rhs)
            if not values:
                logger.info(
                    "ScheduleCodeCase.In: not selected (rhs values unavailable)"
                )
                return super().as_sql(compiler, connection)
            combined_q = dm.Q()
            for value in values:
                q = expr.build_q(value)
                if q:
                    combined_q |= q
            if not combined_q:
                logger.info(
                    "ScheduleCodeCase.In: not selected (empty Q for %r)", values
                )
                return super().as_sql(compiler, connection)
            where = combined_q.resolve_expression(
                compiler.query, allow_joins=True, reuse=None, summarize=False
            )
            logger.info("ScheduleCodeCase.In: selected (optimized)")
            return compiler.compile(where)

    lookup_overrides = {"exact": Exact, "in": In}

    def __init__(
        self,
        *args,
        start_ref: str = "start_date",
        end_ref: str = "end_date",
        current_date=None,
        **kwargs,
    ):
        from django.utils import timezone

        current_date = current_date or timezone.now()
        self.start_ref = start_ref
        self.end_ref = end_ref
        self.current_date = current_date

        if not args:
            args = (
                When(
                    dm.Q(**{f"{start_ref}__lte": current_date})
                    & dm.Q(**{f"{end_ref}__gt": current_date}),
                    then=Value(ScheduleCode.ACTIVE.value),
                ),
                When(
                    **{f"{start_ref}__gt": current_date},
                    then=Value(ScheduleCode.FUTURE.value),
                ),
            )
            kwargs.setdefault("default", Value(ScheduleCode.EXPIRED.value))

        kwargs.setdefault("output_field", dm.CharField())
        super().__init__(*args, **kwargs)

    def build_q(self, code_value):
        """Build Q object for a given schedule code value."""
        if code_value == ScheduleCode.ACTIVE.value:
            # start_date <= current_date < end_date
            return dm.Q(**{f"{self.start_ref}__lte": self.current_date}) & dm.Q(
                **{f"{self.end_ref}__gt": self.current_date}
            )
        if code_value == ScheduleCode.FUTURE.value:
            # start_date > current_date
            return dm.Q(**{f"{self.start_ref}__gt": self.current_date})
        if code_value == ScheduleCode.EXPIRED.value:
            # end_date <= current_date
            return dm.Q(**{f"{self.end_ref}__lte": self.current_date})
        # Unknown code, return empty Q
        return dm.Q()


MIN_DATE = datetime.datetime(1900, 1, 1, tzinfo=pytz.UTC)
MAX_DATE = datetime.datetime(9000, 1, 1, tzinfo=pytz.UTC)


def nest_field(*, ref: str, field_name: str):
    ref = f"{ref}__" if ref else ""
    ref = f"{ref}{field_name}"
    return ref


nest_start_date = partial(nest_field, field_name="start_date")
nest_end_date = partial(nest_field, field_name="end_date")
nest_period = partial(nest_field, field_name="period")


class ScheduleMixin(dm.Model):
    """Base mixin for models with start_date and end_date scheduling fields."""

    start_date = dm.DateTimeField(null=True, db_index=True)
    end_date = dm.DateTimeField(null=True, db_index=True)

    @property
    @abc.abstractmethod
    def objects(self) -> "ScheduleQuerySet":
        return NotImplemented

    @staticmethod
    def get_constraint_end_time_gt_start_time():
        return dm.CheckConstraint(
            condition=dm.Q(end_date__gt=dm.F("start_date")),
            name="%(class)s_end_after_start",
        )

    class Meta:
        abstract = True
        ordering = ["start_date", "end_date"]


class PeriodScheduleMixin(ScheduleMixin):
    """Extended schedule mixin that includes a generated period field for overlap detection."""

    period = dm.GeneratedField(
        expression=dm.Func(
            dm.F("start_date"),
            dm.F("end_date"),
            dm.Value("[)"),  # lower-inclusive / upper-exclusive
            function="tstzrange",
        ),
        output_field=DateTimeRangeField(),
        db_persist=True,  # materialised (required on Postgres)
        editable=False,
    )
    overlap_constraint_fields: tuple[str, ...] = ()

    @staticmethod
    def get_constraint_no_period_overlap_on_fields(field_names: list[str] = None):
        field_names = field_names or []
        return ExclusionConstraint(
            name="%(class)s_no_overlap_" + "_".join(field_names),
            expressions=[
                *[(f, "=") for f in field_names],
                ("period", RangeOperators.OVERLAPS),
            ],
            deferrable=dm.Deferrable.DEFERRED,
        )

    @staticmethod
    def get_default_indexes(
        prefix_fields: list[str] | None = None, class_name_abbr: str | None = None
    ) -> list[dm.Index]:
        """
        Return the 'standard' indexes for temporal-membership tables with period field.

        * `prefix_fields` - equality columns evaluated before the range test.
        * `class_name_abbr` - optional abbreviated class name for index names (to stay within 30-char limit)
        """
        prefix_fields = prefix_fields or []
        name_template = class_name_abbr if class_name_abbr else "%(class)s"

        #  Multi-column GiST on (<prefix ...>, period)
        gist_idx = GistIndex(
            fields=[*prefix_fields, "period"],
            name=f"{name_template}_gist_" + "_".join(prefix_fields + ["period"]),
        )

        # B-tree on upper(period) for 'ending soon' queries
        upper_idx = dm.Index(
            dm.Func(dm.F("period"), function="upper"),
            name=f"{name_template}_idx_period_upper",
        )

        # B-tree on lower(period) for expired lookups
        lower_idx = dm.Index(
            dm.Func(dm.F("period"), function="lower"),
            name=f"{name_template}_idx_period_lower",
        )

        return [gist_idx, upper_idx, lower_idx]

    class Meta:
        abstract = True
        ordering = ["start_date", "end_date"]



# Generated by Django 5.1.6 on 2025-02-22 15:14

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='FeatureFlag',
            fields=[
                ('id', models.CharField(max_length=50, primary_key=True, serialize=False, unique=True)),
                ('active', models.BooleanField(default=False)),
                ('value', models.IntegerField(default=0)),
                ('value_str', models.CharField(default='', max_length=100)),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



import csv
import datetime
import io
import os

from django.core.files.storage import FileSystemStorage, default_storage
from django.db.models import QuerySet
from django.utils.crypto import get_random_string

from django_infra.exporter.models import ExportMetadata, ExportState, QueryExport

EXPORT_BATCH_SIZE = 10_000


def generate_export_id(qs: QuerySet) -> str:
    cls_name = qs.model.__name__
    now = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
    unique_component = get_random_string(6)
    return f"{cls_name}-{now}-{unique_component}"


def handle_csv(fields, rows, file_obj):
    if not hasattr(file_obj, "encoding"):
        file_obj = io.TextIOWrapper(file_obj, encoding="utf-8", newline="")
    writer = csv.writer(file_obj)
    writer.writerow(fields)
    for row in rows:
        writer.writerow(row)
    file_obj.flush()


def export_queryset(qs: QuerySet, values: list, file_path: str = None) -> QueryExport:
    export_id = generate_export_id(qs)
    file_path = file_path or f"exports/{export_id}.csv"
    ext = os.path.splitext(file_path)[1].lower().lstrip(".")
    if not ext:
        raise ValueError("File extension missing")
    if default_storage.exists(file_path):
        raise ValueError("File already exists")

    export = QueryExport.objects.create(
        id=export_id,
        state=ExportState.SCHEDULED,
        metadata=ExportMetadata().data,
        format=ext,
    )
    try:
        export.update(state=ExportState.PROCESSING)
        total = qs.count()
        processed = 0

        def row_generator():
            nonlocal processed
            for row in qs.values_list(*values).iterator(chunk_size=EXPORT_BATCH_SIZE):
                processed += 1
                yield row
                if processed % EXPORT_BATCH_SIZE == 0 and total:
                    export.export_metadata.update_progress(processed, total)
                    export.update(metadata=export.export_metadata.data)

        handler = {"csv": handle_csv}.get(ext)
        if handler is None:
            raise ValueError("Unsupported format")

        if isinstance(default_storage, FileSystemStorage):
            os.makedirs(os.path.dirname(file_path), exist_ok=True)

        with default_storage.open(file_path, "wb") as f:
            handler(values, row_generator(), f)

        file_size = default_storage.size(file_path)
        export.export_metadata.finalize(file_size, processed)
        export.update(state=ExportState.SUCCESS, metadata=export.export_metadata.data)
    except Exception as e:
        export.export_metadata.error_log = str(e)
        export.update(state=ExportState.FAIL, metadata=export.export_metadata.data)
        raise
    return export



from __future__ import annotations

import datetime

from django.contrib.postgres.fields import DateTimeRangeField
from django.db import models as dm
from django.db.models.expressions import RawSQL
from django.utils import timezone

from django_infra.db.models.schedule import (
    ScheduleCodeCase,
    nest_end_date,
    nest_period,
    nest_start_date,
)


class ScheduleQuerySet(dm.QuerySet):
    def filter(self, *args, **kwargs) -> ScheduleQuerySet:
        return super().filter(*args, **kwargs)

    @staticmethod
    def annotate_period(*, qs, start_ref="start_date", end_ref="end_date", to="period"):
        """
        Annotate the queryset with a `tstzrange` built from `start_ref` -> `end_ref`.
        If the start timestamp is after the end timestamp, return an *empty* range
        instead of letting Postgres raise an error.
        """

        range_expr = dm.Func(
            dm.F(start_ref),
            dm.F(end_ref),
            dm.Value("[)"),  # inclusive lower, exclusive upper
            function="tstzrange",
            output_field=DateTimeRangeField(),
        )

        empty_range = RawSQL(
            "%s::tstzrange",
            params=["empty"],
            output_field=DateTimeRangeField(),
        )

        annotated_range = dm.Case(
            dm.When(**{f"{start_ref}__lte": dm.F(end_ref)}, then=range_expr),
            default=empty_range,
        )

        return qs.annotate(**{to: annotated_range})

    @staticmethod
    def annotate_within_schedule(
        *,
        qs,
        date_time=None,
        schedule_ref: str = "",
        annotation_name="within_schedule",
    ):
        current_date = date_time or timezone.now()

        # Use start_date and end_date directly (faster than period field)
        start_ref = nest_start_date(ref=schedule_ref)
        end_ref = nest_end_date(ref=schedule_ref)

        # Use custom ScheduleCodeCase that optimizes filtering
        code_case = ScheduleCodeCase(
            start_ref=start_ref,
            end_ref=end_ref,
            current_date=current_date,
        )

        return qs.annotate(
            **{
                f"{annotation_name}_code": code_case,
                annotation_name: dm.ExpressionWrapper(
                    dm.Q(**{f"{start_ref}__lte": current_date})
                    & dm.Q(**{f"{end_ref}__gt": current_date}),
                    output_field=dm.BooleanField(),
                ),
            }
        )

    def with_within_schedule(self, date_time=None, annotation_name="within_schedule"):
        """Annotate self with a parameter stating if date is within schedule range."""
        return self.annotate_within_schedule(
            qs=self, date_time=date_time, annotation_name=annotation_name
        )


class PeriodScheduleQuerySet(ScheduleQuerySet):
    """Extended queryset for models with period field, adding overlap detection methods."""

    @staticmethod
    def annotate_has_overlap(*, qs, schedule_ref: str = "", partition_by: str = ""):
        period_ref = nest_period(ref=schedule_ref)

        subquery = qs.filter(
            ~dm.Q(pk=dm.OuterRef("pk")),
            **{f"{period_ref}__overlap": dm.OuterRef(period_ref)},
        )
        if partition_by:
            subquery = subquery.filter(**{partition_by: dm.OuterRef(partition_by)})

        return qs.annotate(has_overlap=dm.Exists(subquery))

    def with_has_overlap(self):
        return PeriodScheduleQuerySet.annotate_has_overlap(qs=self)

    @staticmethod
    def annotate_range_has_overlap(
        *,
        qs,
        start: str | datetime.datetime,
        end: str | datetime.datetime,
        schedule_ref: str = "",
        annotation_name: str = "range_has_overlap",
    ):
        period_ref = nest_period(ref=schedule_ref)
        overlap_key = f"{period_ref}__overlap"

        return qs.annotate(
            **{
                annotation_name: dm.Case(
                    dm.When(**{overlap_key: (start, end)}, then=dm.Value(True)),
                    default=dm.Value(False),
                    output_field=dm.BooleanField(),
                )
            }
        )

    def with_range_has_overlap(
        self, start: str | datetime.datetime, end: str | datetime.datetime
    ):
        """Annotate self with a parameter stating if date is within schedule range."""
        return self.annotate_range_has_overlap(qs=self, start=start, end=end)



from __future__ import annotations

import re
import typing
from typing import List, Type, Union

from django.db import models as dm
from django_filters import rest_framework as filters


def field_supports_partial_matching(model, field_name):
    """True if field (nested or not) supports partial matching.

    Fields supporting partial matching:
    - TextField
    - CharField

    Examples
    ---------
    >>> field_supports_partial_matching(PatientJourney,'patient__user__first_name')
    >>> True
    >>> field_supports_partial_matching(PatientJourney,'patient__user_id')
    >>> False
    >>> field_supports_partial_matching(PatientJourney,'id')
    >>> False
    """
    parts = field_name.split("__")
    field = None
    for part in parts:
        if field is None:
            # Start from the base model
            field = model._meta.get_field(part)
        else:
            # Traverse related fields
            if hasattr(field, "remote_field") and field.remote_field:
                model = field.remote_field.model
                field = model._meta.get_field(part)
            else:
                # Field cannot be traversed further
                return False
    return isinstance(field, (dm.CharField, dm.TextField))


class FilterField:
    internal: str = None
    external: str = None
    exclude: bool = False

    def __init__(
        self, *, internal: dm.Field | str, external: str = None, exclude=False
    ):
        """
        internal: variable used in orm e.g.: user__first_name
        external: variable used by client: patient_name
        """
        self.exclude = exclude
        if self.internal is None and internal is None:
            raise RuntimeError("Internal not set.")
        if self.internal is None:
            internal = self.normalise_internal(internal)
            self.internal = internal
        if self.external is None:
            self.external = external or internal

    @staticmethod
    def normalise_internal(field: dm.Field | str) -> str:
        if hasattr(field, "field"):
            field = field.field.attname
        return field

    @property
    def field(self) -> tuple[str, str]:
        return self.internal, self.external

    def retrieve_value_from_instance(self, instance):
        """Retrieve internal field value from instance.

        If the field does have the internal attribute we split by `__`
        and attempt a nested retrieval

        Raises: AttributeError
        """
        if hasattr(instance, self.internal):
            return getattr(instance, self.internal)
        field_tree = self.internal.split("__")
        if len(field_tree) == 1:
            return getattr(instance, self.internal)  # will raise
        # try to perform a nested retrieval
        # e.g. pj.team.lead.user.first_name
        # given "team__lead__user__first_name"
        value = instance
        while field_tree and value is not None:
            node = field_tree.pop(0)
            try:
                value = getattr(value, node)
            except AttributeError as e:
                raise AttributeError(
                    f"Failed to get {self.internal} from {instance}: {e}"
                )
        return value


class TypedFilterField(FilterField):
    type_: Type[bool | int | str | float]

    def __init__(
        self,
        *,
        type_: Type[bool | int | str | float],
        internal: Union[dm.Field, str],
        external: str = None,
        exclude: bool = False,
    ):
        self.type_ = type_
        super().__init__(internal=internal, external=external, exclude=exclude)


class MultiInternalFilterField(FilterField):
    internal: List[str]

    def __init__(
        self,
        *,
        internal: List[Union[dm.Field, str]],
        external: str,
        exclude: bool = False,
    ):
        self.internal = [self.normalise_internal(f) for f in internal]
        self.external = external
        super().__init__(internal="", exclude=exclude)


FIELD_TYPES = FilterField | MultiInternalFilterField | TypedFilterField


class Filter:
    fields = List[FIELD_TYPES]
    _internal_map: dict[str, FilterField]
    _external_map: dict[str, FilterField]
    name: str

    def __init__(
        self,
        *,
        fields: List[FilterField | MultiInternalFilterField | TypedFilterField],
        name: str,
    ):
        self.fields = fields
        self.name = name
        self._validate()
        self.fields = fields

    def _validate(self):
        self._external_map = dict()
        self._internal_map = dict()
        if self.fields is None:
            raise ValueError("fields should not be None")
        for f in self.fields:
            if not isinstance(f, FilterField):
                raise RuntimeError(f"{f} should be of type {FilterField}")
            self._external_map[f.external] = f
            if not isinstance(f, MultiInternalFilterField):
                self._internal_map[f.internal] = f

    def add_field(self, field: FIELD_TYPES) -> Filter:
        self.fields.append(field)
        self._validate()
        return self

    @property
    def fields_names(self):
        return [f.field for f in self.fields]


class CustomFilter(Filter):
    fields: List[TypedFilterField]

    def __init__(self, *, fields, custom_filter: typing.Callable):
        """Partial matching of string into one or more internal fields
        Examples
        ---------
        >>> CustomFilter(
        >>>     custom_filter:typing.Callable=<my_filter>
        >>> )
        >>> client.get(url="/?search_by_name=<name>")
        >>> client.get(url="/?search_by_slug=<slug>")
        Parameters
        ----------
        fields:List[MultiInternalFilterField]
        """
        super().__init__(fields=fields, name="icontains_filter")
        self.custom_filter = custom_filter

    def get_filter(self, qs, external_field: str, payload_value: typing.Any):
        """Returns custom filter implementation with external map injected into kwargs.
        !!! Attention, there is some gimmicky stuff going on here !!!
        Django filters requires you to add a `self` attr when defining a custom filter
        So essentially here `self` is both acting as the instance of this class
        AND a reference to the django filter.

        What is important is that we clean this data up when handing it off to
        the custom filter function specified, (we don't pass any `self` instance
        to avoid this confusion.
        """
        return self.custom_filter(
            qs=qs, value=payload_value, external_map=self._external_map
        )

    def to_meta(self):
        # this is what is used in the query see filter fn.
        d = dict()
        for f in self.fields:
            d.update(
                {
                    f"{f.external}": filters.Filter(
                        field_name=f.external,
                        method=f"filter_{f.external}",
                    ),
                    f"filter_{f.external}": self.get_filter,
                }
            )
        return d


class _NullLastOrderingFilter(filters.OrderingFilter):
    def get_ordering_value(self, param):
        descending = param.startswith("-")
        param = param[1:] if descending else param
        field_name = self.param_map.get(param, param)
        if descending:
            return dm.F(field_name).desc(nulls_last=True)
        else:
            return dm.F(field_name).asc(nulls_last=True)


class OrderingFilter(Filter):
    def __init__(
        self, *, fields: List[FilterField], name: str, base_ordering_filter=None
    ):
        """Orders by one or more client defined string mapped to internal fields
        Examples
        ---------
        >>> OrderingFilter(
        >>>     name='ordering',
        >>>     fields=[
        >>>         FilterField(
        >>>             internal='team__lead_id',
        >>>             external='order_by_lead_id',
        >>>         ),
        >>>         FilterField(
        >>>             internal='slug',
        >>>             external='order_by_slug',
        >>>         ),
        >>>     ]
        >>> )
        >>> client.get(url="/?ordering=-order_by_slug,order_by_lead_id")

        Parameters
        ----------
        fields:List[FilterField]
        """
        self.base_ordering_filter = base_ordering_filter or _NullLastOrderingFilter
        super().__init__(fields=fields, name=name)

    def to_meta(self):
        return {
            self.name: self.base_ordering_filter(
                field_name=self.name, fields=self.fields_names
            )
        }


class IContainsFilter(Filter):
    fields: List[MultiInternalFilterField]

    @staticmethod
    def format_regex(value: str) -> str:
        """
        Creates an ORM readable regex string, escaping all initial
        regex characters except * which is used as wildcard
        """
        value = re.escape(value)
        value = f"^{value}$"
        value = value.replace("\\*", ".*")
        return value

    def __init__(self, *, fields: List[MultiInternalFilterField], allow_wildcard=False):
        """Partial matching of string into one or more internal fields
        Examples
        ---------
        >>> IContainsFilter(
        >>>     fields=[
        >>>         MultiInternalFilterField(
        >>>             internal=['team__lead__user__first_name',
        'patient__user__first_name'],
        >>>             external='search_by_name')
        >>>         ),
        >>>         MultiInternalFilterField(
        >>>             internal=['journey__slug','slug'],
        >>>             external='search_by_slug')
        >>>         ),
        >>>     ]
        >>> )
        >>> client.get(url="/?search_by_name=<name>")
        >>> client.get(url="/?search_by_slug=<slug>")

        Parameters
        ----------
        fields:List[MultiInternalFilterField]
        """
        super().__init__(fields=fields, name="icontains_filter")
        self.allow_wildcard = allow_wildcard

    def to_meta(self):
        # this is what is used in the query see filter fn.
        d = dict()
        external_map = self._external_map
        for f in self.fields:

            def filter(self_, queryset, name, value):
                q_objects = dm.Q()
                for internal in external_map.get(name).internal:
                    fn = "icontains"
                    if "*" in value and self.allow_wildcard:
                        fn = "iregex"
                        value = self.format_regex(value)
                    if field_supports_partial_matching(queryset.model, internal):
                        q_objects |= dm.Q(**{f"{internal}__{fn}": value})
                    elif value.isnumeric():
                        # field does not support partial matching,
                        # most likely FK & val needs to be numeric.
                        q_objects |= dm.Q(**{internal: value})
                return queryset.filter(q_objects)

            d.update(
                {
                    f"{f.external}": filters.CharFilter(
                        # this is what is used in the url
                        field_name=f.external,
                        method=f"filter_{f.external}",
                    ),
                    f"filter_{f.external}": filter,
                }
            )
        return d


def format_regex(value: str) -> str:
    """
    Creates an ORM readable regex string, escaping all initial
    regex characters except * which is used as wildcard
    """
    value = re.escape(value)
    value = f"^{value}$"

    value = value.replace("\\*", ".*")

    return value


class WildcardFilter(filters.CharFilter):
    def filter(self, qs, value):
        """
        Custom char filter which accepts a string of substring parts to match against,
        optionally separated by one or more wildcards, denoted with '*'

        Args:
            qs (QuerySet): The full queryset passed from the view before filtering
            value (str): The value to filter against

        Example values:
        - exact match: foofoofoo
        - starts with match: *foo
        - ends with match: foo*
        - multiple wildcard separators: f*f*f*

        Returns:
            QuerySet
        """

        # If filter value is empty, return unfiltered
        if not value:
            return qs

        wildcard = "*"

        # No wildcard - return exact case insensitive match
        if wildcard not in value:
            return qs.filter(**{f"{self.field_name}__iexact": value})

        value = format_regex(value)

        qs = qs.filter(**{f"{self.field_name}__iregex": value})

        return qs


class ExactMatchFilter(Filter):
    fields: List[TypedFilterField]

    def __init__(self, *, fields: List[TypedFilterField]):
        """Exact matching of string into internal field
        Examples
        ---------
        >>> ExactMatchFilter(
        >>>     fields=[
        >>>         TypedFilterField(
        >>>             internal='team__lead_id',
        >>>             external='match_on_lead_id',
        >>>             type_=int,
        >>>         ),
        >>>         TypedFilterField(
        >>>             internal='slug',
        >>>             external='match_on_slug',
        >>>             type_=str,
        >>>         ),
        >>>     ]
        >>> )
        >>> client.get(url="/?match_on_lead_id=<id:int>")
        >>> client.get(url="/?match_on_slug=<slug:str>")

        Parameters
        ----------
        fields:List[TypedFilterField]
        """
        super().__init__(fields=fields, name="exact_match_filter")

    @staticmethod
    def get_filter(type_):
        return {
            str: filters.CharFilter,
            bool: filters.BooleanFilter,
            int: filters.NumberFilter,
            float: filters.NumberFilter,
            typing.Any: WildcardFilter,
        }.get(type_)

    def to_meta(self):
        # this is what is used in the query see filter fn.
        d = dict()
        external_map = self._external_map
        for f in self.fields:

            def filter(self, queryset, name, value):
                return queryset.filter(**{f"{external_map.get(name).internal}": value})

            d.update(
                {
                    f"{f.external}": self.get_filter(f.type_)(
                        # this is what is used in the url
                        field_name=f.external,
                        method=f"filter_{f.external}",
                    ),
                    f"filter_{f.external}": filter,
                }
            )
        return d


class StringInFilter(filters.BaseInFilter, filters.CharFilter):
    """
    Comma separated string filter.

    Can be used for `in` or `overlap` lookups for strings and arrays.

    """

    pass


class MatchesAnyOfFilter(Filter):

    def __init__(self, *, fields: List[FilterField]):
        """Exact matching of multiple strings into internal field
        Examples
        ---------
        >>> MatchesAnyOfFilter(
        >>>     fields=[
        >>>         FilterField(
        >>>             internal='team__lead_id',
        >>>             external='match_on_lead_id',
        >>>         ),
        >>>         FilterField(
        >>>             internal='slug',
        >>>             external='match_on_slug',
        >>>         ),
        >>>     ]
        >>> )
        >>> client.get(url="/?match_on_lead_id=<id1:int>,<id2:int>,<id3:int>")
        >>> client.get(url="/?match_on_slug=<slug1:str>,<slug2:str>,<slug3:str>")

        Parameters
        ----------
        fields:List[FilterField]
        """
        super().__init__(fields=fields, name="matches_any_of")

    def to_meta(self):
        # this is what is used in the query see filter fn.
        d = dict()
        for f in self.fields:
            d.update(
                {
                    f"{f.external}": StringInFilter(
                        field_name=f.internal,
                        help_text="Filter by comma separated contact values",
                        exclude=f.exclude,
                    ),
                }
            )
        return d



from django.contrib import admin

from .models import FeatureFlag


@admin.register(FeatureFlag)
class FeatureFlagAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "active",
        "value",
    )
    list_filter = ("active",)
    search_fields = ("id",)
    actions = ["activate_flags", "deactivate_flags"]

    @admin.action(description="Activate selected feature flags")
    def activate_flags(self, request, queryset):
        queryset.update(active=True)

    @admin.action(description="Deactivate selected feature flags")
    def deactivate_flags(self, request, queryset):
        queryset.update(active=False)
